import pytest
import torch
from tasks import get_conv_gn_task, to_device
from utils import check_tensors

from nngeometry.object.vector import PVector, random_pvector
from nngeometry.utils import grad


@pytest.fixture(autouse=True)
def make_test_deterministic():
    torch.manual_seed(1234)
    yield


def test_grad_dict_repr():
    loader, lc, parameters, model, function = get_conv_gn_task()

    d, _ = next(iter(loader))
    scalar_output = model(to_device(d)).sum()
    vec = PVector.from_model(model)

    grad_nng = grad(scalar_output, vec, retain_graph=True)

    scalar_output.backward()
    grad_direct = PVector.from_model_grad(model)

    check_tensors(
        grad_direct.to_torch(), grad_nng.to_torch()
    )


def test_grad_flat_repr():
    loader, lc, parameters, model, function = get_conv_gn_task()

    vec = random_pvector(lc)
    scalar_output = vec.norm()

    with pytest.raises(RuntimeError):
        grad(scalar_output, vec)
